import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
from tensorflow.keras.activations import relu, sigmoid
from tensorflow.keras.initializers import orthogonal, he_uniform
from tensorflow.python.keras.engine.keras_tensor import KerasTensor

from TSFEDL.blocks_keras import densenet_transition_block, densenet_dense_block, squeeze_excitation_module, \
    RTA_block, spatial_attention_block_ZhangJin, temporal_attention_block_ZhangJin
from TSFEDL.utils import check_inputs, full_convolution


def OhShuLih(include_top=True,
             weights=None,
             input_tensor=None,
             input_shape=None,
             classes=5,
             classifier_activation="softmax"):
    """
    CNN+LSTM model for Arrythmia classification.

    Parameters
    ----------
        include_top: bool, default=True
          Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
    `keras.Model`
        A `keras.Model` instance.

    References
    ----------
        `Oh, Shu Lih, et al. "Automated diagnosis of arrhythmia using combination of CNN and LSTM techniques with
        variable length heart beats." Computers in biology and medicine 102 (2018): 278-287.`
    """

    # Initial checks
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model Definition
    x = full_convolution(inp, filters=3, kernel_size=20, activation=relu, use_bias=False, strides=1)
    x = layers.MaxPooling1D(pool_size=2)(x)
    x = full_convolution(x, filters=6, kernel_size=10, activation=relu, use_bias=False, strides=1)
    x = layers.MaxPooling1D(pool_size=2)(x)
    x = full_convolution(x, filters=6, kernel_size=5, activation=relu, use_bias=False, strides=1)
    x = layers.MaxPooling1D(pool_size=2)(x)
    x = layers.LSTM(units=20, recurrent_dropout=0.2)(x)

    if include_top:
        x = layers.Dropout(rate=0.2)(x)
        x = layers.Dense(units=20, activation=relu)(x)
        x = layers.Dense(units=10, activation=relu)(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="OhShuLih")

    # Load weights
    if weights is not None:
        model.load_weights(weights)

    return model


def KhanZulfiqar(include_top=True,
                 weights=None,
                 input_tensor=None,
                 input_shape=None,
                 gru_units=(100, 50),
                 return_sequences=False,
                 classes=5,
                 classifier_activation="softmax"):
    """
    CNN+GRU model for electricity forecasting

    Parameters
    ----------
        include_top: bool, default=True
          Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        gru_units: Tuple of length=2, defaults=(100, 50)
            The number of units within the 2 GRU layers.

        return_sequences: bool, defaults=False
            If True, the last GRU layer and Top layer if `include_top=True` returns the whole sequence instead of
            the last value.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
        model: keras.Model
            A `keras.Model` instance with the Full CNN-LSTM Autoencoder

    References
    ---------
        Sajjad, M., Khan, Z. A., Ullah, A., Hussain, T., Ullah, W., Lee, M. Y., & Baik, S. W. (2020). A novel
        CNN-GRU-based hybrid approach for short-term residential load forecasting. IEEE Access, 8, 143759-143768.
     """

    # Check inputs
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)
    if len(gru_units) != 2:
        raise ValueError("'gru_units' must be a tuple of length 2")

    # Model Definition
    x = layers.Conv1D(filters=16, kernel_size=2, activation=relu)(inp)
    x = layers.Dropout(0.1)(x)
    x = layers.Conv1D(filters=8, kernel_size=2, activation=relu)(x)
    x = layers.Dropout(0.1)(x)
    x = layers.GRU(units=gru_units[0], return_sequences=True)(x)
    x = layers.GRU(units=gru_units[1], return_sequences=return_sequences)(x)

    if include_top:
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="KhanZulfiqar")

    if weights is not None:
        model.load_weights(weights)

    return model


def ZhengZhenyu(include_top=True,
                weights=None,
                input_tensor=None,
                input_shape=None,
                classes=5,
                classifier_activation="softmax"):
    """

    CNN+LSTM network for arrythmia detection. This model initialy was designed to deal with 2D images. It was adapted
    to 1D time series.

    Parameters
    ----------
        include_top: bool, default=True
            Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
        model: keras.Model
            A `keras.Model` instance.

    References
    ----------
        Zheng, Z., Chen, Z., Hu, F., Zhu, J., Tang, Q., & Liang, Y. (2020). An automatic diagnosis of arrhythmias using
        a combination of CNN and LSTM technology. Electronics, 9(1), 121.
    """
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model Definition
    x = inp
    for f in [64, 128, 256]:
        x = layers.Conv1D(filters=f, kernel_size=3, strides=1, padding='same', activation=keras.activations.elu)(x)
        x = layers.BatchNormalization()(x)
        x = layers.Conv1D(filters=f, kernel_size=3, strides=1, padding='same', activation=keras.activations.elu)(x)
        x = layers.BatchNormalization()(x)
        x = layers.MaxPooling1D(pool_size=2, strides=2)(x)

    x = layers.LSTM(units=256)(x)

    if include_top:
        x = layers.Dense(units=2048, activation=keras.activations.elu)(x)
        x = layers.Dropout(rate=0.5)(x)
        x = layers.Dense(units=2048, activation=keras.activations.elu)(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="ZhengZhenyu")

    # Load weights
    if weights is not None:
        model.load_weights(weights)

    return model


def HouBoroui(weights=None,
              input_tensor=None,
              input_shape=None,
              encoder_units=100):
    """
    Basic LSTM Autoencoder, where the encoded features were introduced in an SVM.

    Parameters
    ----------
        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        encoder_units: int, defaults=100
            The number of encoding features.

    Returns
    -------
        autoencoder: keras.Model
            A `keras.Model` instance representing the full autoencoder.
        encoder: keras.Model
            A `keras.Model` instance that representes the encoder.

    References
    ----------
        Hou, Borui, et al. "LSTM based auto-encoder model for ECG arrhythmias classification." IEEE Transactions on
        Instrumentation and Measurement (2019).

    """
    inp = check_inputs(True, weights, input_tensor, input_shape, 5, None)  # only check weights and inputs.

    # Model definition
    encoder = layers.LSTM(units=encoder_units)(inp)
    x = layers.RepeatVector(inp.shape[1])(encoder)
    x = layers.LSTM(units=encoder_units, return_sequences=True)(x)
    x = layers.TimeDistributed(layers.Dense(units=1))(x)

    autoencoder = keras.Model(inputs=inp, outputs=x, name="HouBoroui")

    if weights is not None:
        autoencoder.load_weights(weights)

    encoder = keras.Model(inputs=inp, outputs=encoder)

    return autoencoder, encoder


def WangKejun(include_top=True,
              weights=None,
              input_tensor=None,
              input_shape=None,
              classes=5,
              classifier_activation="softmax"):
    """
    References
    ----------
        Wang, Kejun, Xiaoxia Qi, and Hongda Liu. "Photovoltaic power forecasting based LSTM-Convolutional Network."
        Energy 189 (2019): 116225.

    Parameters
    ----------
        include_top: bool, default=True
          Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
        model: keras.Model
            A `keras.Model` instance.
    """
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model definition
    x = layers.LSTM(units=64, return_sequences=True)(inp)
    x = layers.LSTM(units=128, return_sequences=True)(x)
    x = layers.Conv1D(filters=64, kernel_size=3, strides=1, activation=relu)(x)
    x = layers.MaxPooling1D(pool_size=2, strides=2)(x)
    x = layers.Conv1D(filters=128, kernel_size=3, strides=1, activation=relu)(x)
    x = layers.MaxPooling1D(pool_size=2, strides=2)(x)

    if include_top:
        x = layers.Flatten()(x)
        x = layers.Dropout(rate=0.1)(x)
        x = layers.Dense(units=2048, activation=relu)(x)
        x = layers.Dense(units=1024, activation=relu)(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="WangKejun")

    if weights is not None:
        model.load_weights(weights)

    return model


def ChenChen(include_top=True,
             weights=None,
             input_tensor=None,
             input_shape=None,
             classes=5,
             classifier_activation="softmax"):
    """
    CNN+LSTM model for arrythmia classification.

    Parameters
    ----------
        include_top: bool, default=True
          Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
    model: keras.Model
        A `keras.Model` instance.

    References
    ----------
        Chen, Chen, et al. "Automated arrhythmia classification based on a combination network of CNN and LSTM."
        Biomedical Signal Processing and Control 57 (2020): 101819.
    """
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model definition
    x = inp
    for k, f in zip([251, 150, 100, 81, 61, 14],
                    [5, 5, 10, 20, 20, 10]):
        x = layers.Conv1D(filters=f, kernel_size=k, strides=1, activation=relu)(x)
        x = layers.MaxPooling1D(pool_size=2, strides=2)(x)
    x = layers.LSTM(units=32, return_sequences=True)(x)
    x = layers.LSTM(units=64, return_sequences=True)(x)

    if include_top:
        x = layers.Flatten()(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="ChenChen")

    if weights is not None:
        model.load_weights(weights)

    return model


def KimTaeYoung(include_top=True,
                weights=None,
                input_tensor=None,
                input_shape=None,
                classes=5,
                classifier_activation="softmax"):
    """
    Parameters
    ----------
        include_top: bool, default=True
           Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
    model: keras.Model
        A `keras.Model` instance.

    References
    ----------
        Kim, Tae-Young, and Sung-Bae Cho. "Predicting residential energy consumption using CNN-LSTM neural networks."
        Energy 182 (2019): 72-81.

    Notes
    -----
        In the original paper, the time-series is windowed. Therefore, a TimeDistributed layer is employed before the
        LSTM to traverse all the generated windows. Here, we do not implement this as it is problem-specific.
        Please note that if you need this layer then you should add it manually.
    """
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model definition

    x = layers.Conv1D(filters=64, kernel_size=2, strides=1, activation=relu)(inp)
    x = layers.MaxPooling1D(pool_size=2, strides=1)(x)
    x = layers.Conv1D(filters=64, kernel_size=2, strides=1, activation=relu)(x)
    x = layers.MaxPooling1D(pool_size=2, strides=1)(x)
    x = layers.LSTM(units=64, activation=keras.activations.tanh)(x)

    if include_top:
        x = layers.Dense(units=32, activation=relu)(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="KimTaeYoung")

    if weights is not None:
        model.load_weights(weights)

    return model


def GenMinxing(include_top=True,
               weights=None,
               input_tensor=None,
               input_shape=None,
               classes=5,
               classifier_activation="softmax"):
    """
    Parameters
    ----------
        include_top: bool, default=True
          Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
    model: keras.Model
        A `keras.Model` instance.

    References
    ----------
        Geng, Minxing, et al. "Epileptic Seizure Detection Based on Stockwell Transform and Bidirectional Long
        Short-Term Memory." IEEE Transactions on Neural Systems and Rehabilitation Engineering 28.3 (2020): 573-580.
    """
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model definition
    x = layers.Bidirectional(layers.LSTM(units=40, return_sequences=True))(inp)
    if include_top:
        x = layers.Flatten()(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="GenMixing")

    if weights is not None:
        model.load_weights(weights)

    return model


def FuJiangmeng(include_top=True,
                weights=None,
                input_tensor=None,
                input_shape=None,
                classes=5,
                classifier_activation="softmax"):
    """
    Parameters
    ----------
        include_top: bool, default=True
            Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
    model: keras.Model
        A `keras.Model` instance.

    References
    ----------
        Fu, Jiangmeng, et al. "A hybrid CNN-LSTM model based actuator fault diagnosis for six-rotor UAVs." 2019
        Chinese Control And Decision Conference (CCDC). IEEE, 2019.

    """
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model definition
    x = layers.Conv1D(filters=32, kernel_size=1, padding='same', activation=keras.activations.tanh)(inp)
    x = layers.MaxPooling1D(pool_size=2, strides=1)(x)
    x = layers.LSTM(units=256, activation=keras.activations.tanh, dropout=0.3)(x)

    if include_top:
        x = layers.Dense(units=128, activation=keras.activations.tanh)(x)
        x = layers.Dropout(rate=0.3)(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="FuJiangmeng")

    if weights is not None:
        model.load_weights(weights)

    return model


def ShiHaotian(include_top=True,
               weights=None,
               input_tensor=None,
               input_shape=None,
               classes=5,
               classifier_activation="softmax"):
    """

    Parameters
    ----------
        include_top: bool, default=True
          Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
    model: keras.Model
        A `keras.Model` instance.

    References
    ----------
        Shi, Haotian, et al. "Automated heartbeat classification based on deep neural network with multiple input
        layers." Knowledge-Based Systems 188 (2020): 105036.

    """
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model Definition
    if isinstance(inp, KerasTensor) or inp.shape[-1] == 1:
        x1 = layers.Conv1D(filters=32, kernel_size=13, strides=2)(inp)
        x2 = layers.Conv1D(filters=32, kernel_size=13, strides=1)(inp)
        x3 = layers.Conv1D(filters=32, kernel_size=13, strides=2)(inp)
    elif inp.shape[-1] == 3 or inp.shape[-1] == 4:
        x1 = layers.Conv1D(filters=32, kernel_size=13, strides=2)(inp[0])
        x2 = layers.Conv1D(filters=32, kernel_size=13, strides=1)(inp[1])
        x3 = layers.Conv1D(filters=32, kernel_size=13, strides=2)(inp[2])
    else:
        raise ValueError("input must be a vector of tensors of length 1 or 3.")

    x1 = layers.LeakyReLU()(x1)
    x1 = layers.MaxPooling1D(pool_size=2, strides=2)(x1)

    x2 = layers.LeakyReLU()(x2)
    x2 = layers.MaxPooling1D(pool_size=2, strides=2)(x2)

    x3 = layers.LeakyReLU()(x3)
    x3 = layers.MaxPooling1D(pool_size=2, strides=2)(x3)

    x = layers.Concatenate(axis=1)([x1, x2, x3])
    x = layers.LSTM(units=32, return_sequences=True)(x)

    if include_top:
        x = layers.Flatten()(x)
        x = layers.Dense(units=518, activation=relu)(x)
        x = layers.Dense(units=88, activation=relu)(x)
        if not isinstance(inp, KerasTensor) and inp.shape[-1] == 4:
            x = layers.Concatenate([inp[3], x])
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="ShiHaotian")

    if weights is not None:
        model.load_weights(weights)

    return model


def HuangMeiLing(include_top=True,
                 weights=None,
                 input_tensor=None,
                 input_shape=None,
                 classes=5,
                 classifier_activation="softmax",
                 units_in_dense=(512, 256)):
    """
     CNN model, employed for 2-D images of ECG data. This model is adapted for 1-D time series

    Parameters
    ----------
        include_top: bool, default=True
            Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

        units_in_dense: Tuple[int, int], defaults=(512, 256)
            If `include_top=True`, the number of hidden units of the two dense layers before the softmax. Note:
            The default values are arbitrarely set. The number of units in these layers are not specified in the
            original paper.

    Returns
    -------
    model: keras.Model
        A `keras.Model` instance.

    References
    -------
        Huang, Mei-Ling, and Yan-Sheng Wu. "Classification of atrial fibrillation and normal sinus rhythm based on
        convolutional neural network." Biomedical Engineering Letters (2020): 1-11.
     """
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model definition (It is the first configuration of paper's table 4)
    x = inp
    for p, f, k, s in zip([1, 3],
                       [48, 256],
                       [15, 7],
                       [6, 2]):
        x = layers.ZeroPadding1D(padding=p)(x)
        x = layers.Conv1D(filters=f, kernel_size=k, strides=s, activation=relu)(x)
        x = layers.MaxPooling1D(pool_size=2, strides=1)(x)

    if include_top:
        x = layers.Flatten()(x)
        x = layers.Dropout(rate=0.5)(x)
        x = layers.Dense(units=units_in_dense[0], activation=relu)(x)
        x = layers.Dense(units=units_in_dense[1], activation=relu)(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="HuangMeiLing")

    if weights is not None:
        model.load_weights(weights)

    return model


def LihOhShu(include_top=True,
             weights=None,
             input_tensor=None,
             input_shape=None,
             classes=5,
             classifier_activation="softmax"):
    """
    CNN+LSTM model

    Parameters
    ----------
        include_top: bool, default=True
          Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
    model: keras.Model
        A `keras.Model` instance.

    References
    ----------
        Lih, Oh Shu, et al. "Comprehensive electrocardiographic diagnosis based on deep learning." Artificial
        Intelligence in Medicine 103 (2020): 101789.
    """
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model definition
    x = inp
    for filters, k_size in zip([3, 6, 6, 6, 6],
                               [20, 10, 5, 5, 10]):
        x = layers.Conv1D(filters=filters, activation=relu, kernel_size=k_size, strides=1, use_bias=False)(x)
        x = layers.MaxPooling1D(pool_size=2, strides=2)(x)

    x = layers.LSTM(units=10)(x)

    if include_top:
        # x = layers.Flatten()(x)
        x = layers.Dense(units=8)(x)
        x = layers.Dropout(rate=0.5)(x)
        x = layers.Dense(units=8)(x)
        x = layers.Dropout(rate=0.5)(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="LihOhShu")

    if weights is not None:
        model.load_weights(weights)

    return model


def GaoJunLi(include_top=True,
             weights=None,
             input_tensor=None,
             input_shape=None,
             classes=5,
             classifier_activation="softmax"):
    """
    LSTM network

    Parameters
    ----------
        include_top: bool, default=True
          Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
    model: keras.Model
        A `keras.Model` instance.

    References
    ----------
        Gao, Junli, et al. "An Effective LSTM Recurrent Network to Detect Arrhythmia on Imbalanced ECG Dataset."
        Journal of healthcare engineering 2019 (2019).
    """
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model definition
    x = layers.LSTM(units=64, dropout=0.3)(inp)

    if include_top:
        x = layers.Dense(units=32)(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="GaoJunLi")

    if weights is not None:
        model.load_weights(weights)

    return model


def WeiXiaoyan(include_top=True,
               weights=None,
               input_tensor=None,
               input_shape=None,
               classes=5,
               classifier_activation="softmax"):
    """
     CNN+LSTM model employed for 2-D images of ECG data. This model is adapted for 1-D time series.

    Parameters
    ----------
        include_top: bool, default=True
          Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
    model: keras.Model
        A `keras.Model` instance.

    References
    ----------
        Wei, Xiaoyan, et al. "Early prediction of epileptic seizures using a long-term recurrent convolutional
        network." Journal of neuroscience methods 327 (2019): 108395.
    """
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model definition
    x = inp
    # Network Defintion (CNN)
    for filters, kernel, strides in zip([32, 64, 128, 256, 512],
                                        [5, 3, 3, 3, 3],
                                        [1, 1, 1, 1, 1]):
        x = layers.Conv1D(filters=filters, kernel_size=kernel, strides=strides)(x)
        x = layers.LeakyReLU(alpha=0.01)(x)
        x = layers.MaxPooling1D(pool_size=2, strides=2)(x)
        x = layers.BatchNormalization()(x)

    # LSTM
    x = layers.LSTM(units=512, return_sequences=True)(x)
    x = layers.BatchNormalization()(x)
    x = layers.LSTM(units=512)(x)

    if include_top:
        x = layers.Dense(units=1024)(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="WeiXiaoyan")

    if weights is not None:
        model.load_weights(weights)

    return model


def KongZhengmin(include_top=True,
                 weights=None,
                 input_tensor=None,
                 input_shape=None,
                 classes=5,
                 classifier_activation="softmax",
                 return_sequences=False):
    """
    CNN+LSTM

    Parameters
    ----------
        include_top: bool, default=True
          Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

        return_sequences: bool, defaults=False
            If True, the last LSTM layer will return the whole sequence instead of the last value.

    Returns
    -------
    model: keras.Model
        A `keras.Model` instance.

    References
    ----------
        Kong, Zhengmin, et al. "Convolution and Long Short-Term Memory Hybrid Deep Neural Networks for Remaining
        Useful Life Prognostics." Applied Sciences 9.19 (2019): 4156.
    """
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model definition
    # Note: Number of filters, kernel_sizes, strides, and nº of units in LSTM layers are not specified in the paper.
    x = layers.Conv1D(filters=32, activation=relu, kernel_size=5, strides=1)(inp)
    x = layers.MaxPooling1D(pool_size=2, strides=2)(x)
    x = layers.LSTM(64, return_sequences=True)(x)
    x = layers.LSTM(64, return_sequences=return_sequences)(x)

    if include_top:
        if return_sequences:
            x = layers.Flatten()(x)
        x = layers.Dense(units=50)(x)
        x = layers.Dense(units=50)(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="KongZhengmin")

    if weights is not None:
        model.load_weights(weights)

    return model


def YildirimOzal(include_top=True,
                 autoencoder_weights=None,
                 lstm_weights=None,
                 input_tensor=None,
                 input_shape=None,
                 classes=5,
                 classifier_activation="softmax"):
    """CAE-LSTM

    Parameters
    ----------
        include_top: bool, default=True
          Whether to include the fully-connected layer at the top of the network.

        autoencoder_weights: str, default=None
            The path to the weights file to be loaded for the autoencoder network

        lstm_weights: str, defaults=None
            The path to the weights file to be loaded for the lstm classifier network.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
        autoenconder: keras.Model
            A `keras.Model` instance with the autoencoder.

        encoder: keras.Model
            A `keras.Model` instance with only the encoder part

        model: keras.Model
            A `keras.Model` instance representing the classification model which uses the encoder part.

    References
    ----------
        Yildirim, Ozal, et al. "A new approach for arrhythmia classification using deep coded features and LSTM
        networks." Computer methods and programs in biomedicine 176 (2019): 121-133.
     """
    inp = check_inputs(include_top, None, input_tensor, input_shape, classes, classifier_activation)
    if autoencoder_weights is not None and not tf.io.gfile.exists(autoencoder_weights):
        raise ValueError("'autoencoder_weights' path does not exists: ", autoencoder_weights)
    if lstm_weights is not None and not tf.io.gfile.exists(lstm_weights):
        raise ValueError("'lstm_weights' path does not exists: ", lstm_weights)

    # Model definition
    e = layers.Conv1D(filters=16, kernel_size=5, padding='same', strides=1, activation=relu)(inp)
    e = layers.MaxPooling1D(pool_size=2)(e)
    e = layers.Conv1D(filters=64, kernel_size=5, padding='same', strides=1, activation=relu)(e)
    e = layers.BatchNormalization()(e)
    e = layers.MaxPooling1D(pool_size=2)(e)
    e = layers.Conv1D(filters=32, kernel_size=3, padding='same', strides=1, activation=relu)(e)
    e = layers.Conv1D(filters=1, kernel_size=3, padding='same', strides=1, activation=relu)(e)

    bottleneck = layers.MaxPooling1D(pool_size=2)(e)
    decoder = layers.Conv1D(filters=1, kernel_size=3, padding='same', strides=1, activation=relu)(bottleneck)
    decoder = layers.Conv1D(filters=32, kernel_size=3, padding='same', strides=1, activation=relu)(decoder)
    decoder = layers.UpSampling1D(size=2)(decoder)
    decoder = layers.Conv1D(filters=64, kernel_size=5, padding='same', strides=1, activation=relu)(decoder)
    decoder = layers.UpSampling1D(size=2)(decoder)
    decoder = layers.Conv1D(filters=16, kernel_size=5, padding='same', strides=1, activation=relu)(decoder)
    # Final Dense layer of the autoencoder
    decoder = layers.Flatten()(decoder)
    decoder = layers.Dense(units=inp.shape[1], activation=sigmoid)(decoder)

    autoencoder = keras.Model(inputs=inp, outputs=decoder, name="YildirimOzal_autoencoder")

    if autoencoder_weights is not None:
        autoencoder.load_weights(autoencoder_weights)
    encoder = keras.Model(inputs=inp, outputs=bottleneck, name="YildirimOzal_encoder")

    lstm = layers.LSTM(units=32, return_sequences=True)(bottleneck)
    if include_top:
        lstm = layers.Flatten()(lstm)
        lstm = layers.Dense(units=128, activation=relu)(lstm)
        lstm = layers.Dropout(rate=0.1)(lstm)
        lstm = layers.Dense(units=classes, activation=classifier_activation)(lstm)

    model = keras.Model(inputs=inp, outputs=lstm, name="YildirimOzal_classifier")
    if lstm_weights is not None:
        model.load_weights(lstm_weights)

    return autoencoder, encoder, model


def CaiWenjuan(include_top=True,
               weights=None,
               input_tensor=None,
               input_shape=None,
               classes=5,
               classifier_activation="softmax",
               reduction_ratio=0.6):
    """
    DDNN network presented in:

    Parameters
    ----------
        include_top: bool, default=True
            Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

        reduction_ratio: float, defaults=0.5
            The reduction ratio in the transition block.

    Returns
    -------
    `keras.Model`
        A `keras.Model` instance.

    References
    ----------
        `Cai, Wenjuan, et al. "Accurate detection of atrial fibrillation from 12-lead ECG using deep neural network."
        Computers in biology and medicine 116 (2020): 103378.`
   """
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model definition
    dense_layers = [2, 4, 6, 4]
    x1 = layers.Conv1D(filters=8, kernel_size=1, strides=1, padding='same')(inp)
    x2 = layers.Conv1D(filters=16, kernel_size=3, strides=1, padding='same')(inp)
    x3 = layers.Conv1D(filters=24, kernel_size=5, strides=1, padding='same')(inp)
    x = layers.Concatenate(axis=2)([x1, x2, x3])

    for i, n_blocks in enumerate(dense_layers):  # dense-block construction
        x = squeeze_excitation_module(x, dense_units=32)
        x = densenet_dense_block(x, n_blocks, growth_rate=6, name="conv" + str(i + 1))
        x = squeeze_excitation_module(x, dense_units=32)
        # Last dense block does not have transition block.
        if i < len(dense_layers) - 1:
            x = densenet_transition_block(x, reduction=reduction_ratio, name="transition" + str(i + 1))

    x = layers.GlobalAveragePooling1D()(x)

    if include_top:
        x = layers.Flatten()(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="CaiWenjuan")

    if weights is not None:
        model.load_weights(weights)

    return model


def KimMinGu(include_top=True,
             weights=None,  # Here, 'weights' is a vector of paths for each model in the ensemble
             input_tensor=None,
             input_shape=None,
             classes=5,
             classifier_activation="softmax"):
    """
    CNN ensemble model. The same model is employed n-times in an ensemble. Then, the same model is employed retraining
    using the features extracted in the ensemble.

    Parameters
    ----------
        include_top: bool, default=True
          Whether to include the fully-connected layer at the top of the network.

        weights: array(str), default=None
            An array with the paths of the weights file for each model of the ensemble.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
        ensemble: list[keras.Model]
            A list with six `keras.Model` instances. One for each model of the ensemble.

    References
    ----------
        Kim, M. G., Choi, C., & Pan, S. B. (2020). Ensemble Networks for User Recognition in Various Situations Based on
        Electrocardiogram. IEEE Access, 8, 36527-36535.
   """

    # NOTE: kernel sizes are not specified in the original paper.
    # Initial checks
    inp = check_inputs(include_top, None, input_tensor, input_shape, classes, classifier_activation)
    if weights is not None:
        for w in weights:
            if not tf.io.gfile.exists(w):
                raise ValueError("'weights' path does not exists: ", weights)

    # Begin model definition
    dropout = [0.5, 0.6, 0.6, 0.7, 0.5, 0.7]

    ensemble = []
    for i, d in enumerate(dropout):
        model = layers.Conv1D(filters=32, kernel_size=3, padding="same", activation=relu)(inp)
        model = layers.MaxPooling1D(pool_size=2)(model)
        for f in [64, 128, 256, 512]:
            model = layers.Conv1D(filters=f, kernel_size=3, padding="same", activation=relu)(model)
            model = layers.MaxPooling1D(pool_size=2)(model)

        if include_top:
            model = layers.Flatten()(model)
            model = layers.Dense(units=1028, activation=relu)(model)
            model = layers.Dropout(d)(model)
            model = layers.Dense(units=1028, activation=relu)(model)
            model = layers.Dense(units=classes, activation=classifier_activation)(model)

        m = keras.Model(inputs=inp, outputs=model)
        if weights is not None and weights[i] is not None:
            m.load_weights(weights[i])
        ensemble.append(m)

    return ensemble


def HtetMyetLynn(include_top=True,
                 weights=None,
                 input_tensor=None,
                 input_shape=None,
                 classes=5,
                 classifier_activation="softmax",
                 use_rnn="gru",
                 rnn_units=40):
    """
    Hybrid CNN+Bidirectional RNN

    Parameters
    ----------
        include_top: bool, default=True
            Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

        use_rnn: str, defaults='gru'
            Whether to user a Bi-direccional RNN after the CNN. Options are `'gru'` for a GRU, `'lstm'` for an LSTM or
            `None` to not use the RNN.

        rnn_units: int, defatuls=40
            If `use_rnn` is not `None`, the number of units in the chosen biderectional RNN.

    Returns
    -------
    model: `keras.Model`
        A `keras.Model` instance.

    References
    ----------
        Lynn, H. M., Pan, S. B., & Kim, P. (2019). A deep bidirectional GRU network model for biometric
        electrocardiogram classification based on recurrent neural networks. IEEE Access, 7, 145395-145405.
    """
    # Check inputs
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model definition
    x = inp
    for filt, ker, stride in zip([30, 30, 60, 60],
                                 [5, 2, 5, 2],
                                 [2, 2, 2, 2]):
        x = layers.Conv1D(filters=filt, kernel_size=ker, padding='same')(x)
        x = layers.MaxPooling1D(pool_size=stride)(x)

    # Only use the CNN, or add a Bi-directional GRU or LSTM after the CNN.
    if use_rnn is None:
        if include_top:
            x = layers.Dense(units=40, activation=sigmoid)(x)
            x = layers.Dense(units=classes, activation=classifier_activation)(x)
    elif use_rnn.lower() == 'lstm':
        # Note: Units of RNN are not specified in the original paper
        x = layers.Bidirectional(layers.LSTM(units=rnn_units, dropout=0.2))(x)
    elif use_rnn.lower() == 'gru':
        x = layers.Bidirectional(layers.GRU(units=rnn_units, dropout=0.2))(x)
    else:
        raise ValueError("'use_rnn' parameter should one of: None, lstm or gru. Given: ", use_rnn)

    # Adds classification layer only in the case of use bidrectional layers.
    if use_rnn is not None:
        if include_top:
            x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="HtetMyetLynn")

    if weights is not None:
        model.load_weights(weights)

    return model


def ZhangJin(include_top=True,
             weights=None,
             input_tensor=None,
             input_shape=None,
             classes=5,
             classifier_activation="softmax",
             decrease_ratio=2):
    """
    A CNN+Bi-Directional GRU with a spatio-temporal attention mechanism.

    Parameters
    ----------
        include_top: bool, default=True
            Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

        decrease_ratio: float, defaults=2
            The decrease ratio of the model.

    Returns
    -------
    model: `keras.Model`
        A `keras.Model` instance.

    References
    ----------
        Zhang, J., Liu, A., Gao, M., Chen, X., Zhang, X., & Chen, X. (2020). ECG-based multi-class arrhythmia detection
        using spatio-temporal attention-based convolutional recurrent neural network.
        Artificial Intelligence in Medicine, 106, 101856.

    Notes
    -----
        The time dimension must be at least 1000 as there is a huge reduction on each convolutional block.
    """

    # Check inputs
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)
    x = inp

    # Model definition
    for conv_layers, filters in zip([2, 2, 3, 3, 3],
                                    [64, 128, 256, 256, 256]):  # 5-block layers
        for i in range(conv_layers):
            x = layers.Conv1D(filters=filters, kernel_size=3, padding="same")(x)
            x = layers.BatchNormalization()(x)
            x = layers.Activation(activation=relu)(x)
        x = layers.MaxPooling1D(pool_size=3)(x)
        x = layers.Dropout(rate=0.2)(x)

        # Adds attention module after each convolutional block
        x_spatial = spatial_attention_block_ZhangJin(decrease_ratio, x)
        x = layers.multiply([x_spatial, x])
        x_temporal = temporal_attention_block_ZhangJin(x)
        x = layers.multiply([x_temporal, x])

    x = layers.Bidirectional(layers.GRU(units=12, return_sequences=True))(x)
    x = layers.Dropout(rate=0.2)(x)
    if include_top:
        x = layers.GlobalMaxPool1D()(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="ZhangJin")

    if weights is not None:
        model.load_weights(weights)

    return model


def YaoQihang(include_top=True,
              weights=None,
              input_tensor=None,
              input_shape=None,
              classes=5,
              classifier_activation="softmax"):
    """
    Attention-based time-incremental convolutional neural network (ATI-CNN)

    Parameters
    ----------
        include_top: bool, default=True
            Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
        model: `keras.Model`
            A `keras.Model` instance.

    References
    ----------
        Yao, Q., Wang, R., Fan, X., Liu, J., & Li, Y. (2020). Multi-class Arrhythmia detection from 12-lead varied-length
        ECG using Attention-based Time-Incremental Convolutional Neural Network. Information Fusion, 53, 174-182.
    """
    # Check inputs
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)
    x = inp

    # Model definition
    # Convolutional layers (spatial):
    for conv_layers, filters in zip([2, 2, 3, 3, 3],
                                    [64, 128, 256, 256, 256]):  # 5-block layers
        for i in range(conv_layers):
            x = layers.Conv1D(filters=filters, kernel_size=3, padding="same", kernel_initializer=he_uniform)(x)
            x = layers.BatchNormalization()(x)
            x = layers.Activation(activation=relu)(x)
        x = layers.MaxPooling1D(pool_size=3, strides=3)(x)

    # Temporal layers (2 LSTM layers):
    x = layers.LSTM(units=32, return_sequences=True, dropout=0.2, kernel_initializer=orthogonal)(x)
    x = layers.LSTM(units=32, return_sequences=True, dropout=0.2, kernel_initializer=orthogonal)(x)

    # Attention module:
    if include_top:
        x1 = layers.Dense(units=32, activation=keras.activations.tanh, kernel_initializer=he_uniform)(x)
        x1 = layers.Dense(units=classes, activation=classifier_activation, kernel_initializer=he_uniform)(x1)
        x = tf.reduce_mean(x1, axis=1)

    model = keras.Model(inputs=inp, outputs=x, name="YaoQihang")

    if weights is not None:
        model.load_weights(weights)

    return model


def YiboGao(include_top=True,
            weights=None,
            input_tensor=None,
            input_shape=None,
            classes=5,
            classifier_activation="softmax",
            return_loss=False):
    """
    CNN using RTA blocks for end-to-end attrial fibrilation detection.

    Parameters
    ----------
        include_top: bool, default=True
            Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

        return_loss: bool, defaults=False
            Whether to return the custom loss function (en_loss) employed for training this model.

    Returns
    -------
        model: `keras.Model`
            A `keras.Model` instance.

    References
    ----------
        Gao, Y., Wang, H., & Liu, Z. (2021). An end-to-end atrial fibrillation detection by a novel residual-based
        temporal attention convolutional neural network with exponential nonlinearity loss.
        Knowledge-Based Systems, 212, 106589.

    Notes
    -----
        Code adapted from the original implementation available at:
            https://github.com/o00O00o/RTA-CNN
    """

    def en_loss(y_true, y_pred):  # Custom loss function

        epsilon = 1.e-7
        gamma = float(0.3)

        y_true = tf.cast(y_true, tf.float32)
        y_pred = tf.clip_by_value(y_pred, epsilon, 1. - epsilon)
        pos_pred = tf.pow(-tf.math.log(y_pred), gamma)
        nag_pred = tf.pow(-tf.math.log(1 - y_pred), gamma)
        y_t = tf.multiply(y_true, pos_pred) + tf.multiply(1 - y_true, nag_pred)
        loss = tf.reduce_mean(y_t)

        return loss

    # Model definition
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)
    x = inp

    for fil, ker, pool in zip([16, 32, 64, 64],
                              [32, 16, 9, 9],
                              [4, 4, 2, 2]):
        x = RTA_block(x, fil, ker)
        x = layers.MaxPooling1D(pool)(x)

    x = layers.Dropout(0.6)(x)

    x = RTA_block(x, 128, 3)
    x = layers.MaxPooling1D(2)(x)
    x = RTA_block(x, 128, 3)
    x = layers.MaxPooling1D(2)(x)
    x = layers.Dropout(0.6)(x)

    if include_top:
        x = layers.Flatten()(x)
        x = layers.Dropout(rate=0.7)(x)
        x = layers.Dense(units=100, activation=relu)(x)
        x = layers.Dropout(rate=0.7)(x)
        x = layers.Dense(classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="YiboGao")

    if weights is not None:
        model.load_weights(weights)

    # Return the model and its custom loss function
    if return_loss:
        return model, en_loss
    else:
        return model

# Original CNN-LSTM model
def HongTan(include_top=True,
                weights=None,
                input_tensor=None,
                input_shape=None,
                classes=5,
                classifier_activation="softmax"):
    """

    CNN+LSTM network for arrythmia detection. Application of stacked convolutional and long short-term memory network
    for accurate identification of CAD ECG signals.

    Parameters
    ----------
        include_top: bool, default=True
            Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
        model: keras.Model
            A `keras.Model` instance.

    References
    ----------
        TAN, Jen Hong, et al. Application of stacked convolutional and long short-term memory network for accurate
        identification of CAD ECG signals. Computers in biology and medicine, 2018, vol. 94, p. 19-26.
    """
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    # Model Definition
    x = inp
    x = layers.Conv1D(filters=40, kernel_size=5, strides=1, padding='same', activation=keras.activations.elu)(x)
    x = layers.MaxPooling1D(pool_size=2, strides=2)(x)
    x = layers.Conv1D(filters=32, kernel_size=3, strides=1, padding='same', activation=keras.activations.elu)(x)
    x = layers.MaxPooling1D(pool_size=2, strides=2)(x)

    x = layers.LSTM(units=32, dropout=0.5, recurrent_dropout=0.25, return_sequences=True)(x)
    x = layers.LSTM(units=16, recurrent_dropout=0.25, return_sequences=True)(x)
    x = layers.LSTM(units=4)(x)

    if include_top:
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="HongTan")

    # Load weights
    if weights is not None:
        model.load_weights(weights)

    return model

def SharPar(include_top=True,
             weights=None,
             input_tensor=None,
             input_shape=None,
             classes=5,
             classifier_activation="softmax"):
    """
    A 1-layer CNN + LSTM model. Originally proposed for depression detection.

    Parameters
    ----------
        include_top: bool, default=True
            Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
    model: `keras.Model`
        A `keras.Model` instance.

    References
    ----------
        Sharma, G., Parashar, A., & Joshi, A. M. (2021). DepHNN: a 
        novel hybrid neural network for electroencephalogram 
        (EEG)-based screening of depression. Biomedical signal 
        processing and control, 66, 102393.

    Notes
    -----

    """

    # Check inputs
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    x = inp
    x = layers.Conv1D(filters=64, kernel_size=5, strides=1, activation=keras.activations.relu)(x)
    x = layers.LSTM(units=32, return_sequences=True)(x)
    x = layers.LSTM(units=16, return_sequences=True)(x)
    x = layers.Dense(units=16, activation=keras.activations.relu)(x)
    x = layers.Dense(units=16, activation=keras.activations.relu)(x)

    if include_top:
        x = layers.Flatten()(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="SharPar")

    if weights is not None:
        model.load_weights(weights)

    return model

def DaiXiLi(include_top=True,
             weights=None,
             input_tensor=None,
             input_shape=None,
             classes=5,
             classifier_activation="softmax"):
    """
    Convolutional neural network model for emotion classification using a non-end-to-end 
    training method that combines bottom-, middle-, and top-layer convolution features.

    Parameters
    ----------
        include_top: bool, default=True
            Whether to include the fully-connected layer at the top of the network.

        weights: str, default=None
            The path to the weights file to be loaded.

        input_tensor: keras.Tensor, defaults=None
            Optional Keras tensor (i.e. output of `layers.Input()`) to use as input for the model.

        input_shape: Tuple, defaults=None
            If `input_tensor=None`, a tuple that defines the input shape for the model.

        classes: int, defaults=5
            If `include_top=True`, the number of units in the top layer to classify data.

        classifier_activation: str or callable, defaults='softmax'
            The activation function to use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.

    Returns
    -------
    model: `keras.Model`
        A `keras.Model` instance.

    References
    ----------
        Dai, J., Xi, X., Li, G., & Wang, T. (2022). EEG-Based Emotion 
        Classification Using Improved Cross-Connected Convolutional 
        Neural Network. Brain Sciences, 12(8), 977.

    Notes
    -----

    """

    # Check inputs
    inp = check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation)

    v3 = layers.Conv1D(filters=128, kernel_size=5, strides=1, activation=keras.activations.relu, padding="same")(inp)
    v3 = layers.MaxPooling1D(pool_size=2, strides=2)(v3)

    v1 = layers.Conv1D(filters=64, kernel_size=5, strides=1, activation=keras.activations.relu, padding="same")(v3)
    v1 = layers.MaxPooling1D(pool_size=2, strides=2)(v1)
    v1 = layers.Dense(units=16, activation=keras.activations.relu)(v1)

    v3 = layers.Conv1D(filters=64, kernel_size=5, strides=1, activation=keras.activations.relu, padding="same")(v3)
    v3 = layers.MaxPooling1D(pool_size=2, strides=2)(v3)

    v2 = layers.Conv1D(filters=32, kernel_size=5, strides=1, activation=keras.activations.relu, padding="same")(v3)
    v2 = layers.MaxPooling1D(pool_size=2, strides=2)(v2)
    v2 = layers.Dense(units=16, activation=keras.activations.relu)(v2)

    v3 = layers.Conv1D(filters=32, kernel_size=5, strides=1, activation=keras.activations.relu, padding="same")(v3)
    v3 = layers.MaxPooling1D(pool_size=2, strides=2)(v3)
    v3 = layers.Dense(units=16, activation=keras.activations.relu)(v3)

    x = layers.concatenate([v1, v2, v3], axis=1)

    if include_top:
        x = layers.Flatten()(x)
        x = layers.Dense(units=classes, activation=classifier_activation)(x)

    model = keras.Model(inputs=inp, outputs=x, name="DaiXiLi")

    if weights is not None:
        model.load_weights(weights)

    return model


from .blocks_keras import densenet_transition_block
from .blocks_keras import densenet_conv_block
from .blocks_keras import densenet_dense_block
from .blocks_keras import squeeze_excitation_module
from .blocks_keras import conv_block_YiboGao
from .blocks_keras import attention_branch_YiboGao
from .blocks_keras import RTA_block
from .blocks_keras import spatial_attention_block_ZhangJin
from .blocks_keras import temporal_attention_block_ZhangJin
from .blocks_pytorch import ConvBlockYiboGao
from .blocks_pytorch import AttentionBranchYiboGao
from .blocks_pytorch import RTABlock
from .blocks_pytorch import SqueezeAndExcitationModule
from .blocks_pytorch import DenseNetTransitionBlock
from .blocks_pytorch import DenseNetConvBlock
from .blocks_pytorch import DenseNetDenseBlock
from .blocks_pytorch import SpatialAttentionBlockZhangJin
from .blocks_pytorch import TemporalAttentionBlockZhangJin
from .data import get_mit_bih_segments
from .data import read_mit_bih
from .data import MIT_BIH
from .models_keras import OhShuLih
from .models_keras import KhanZulfiqar
from .models_keras import ZhengZhenyu
from .models_keras import HouBoroui
from .models_keras import WangKejun
from .models_keras import ChenChen
from .models_keras import KimTaeYoung
from .models_keras import GenMinxing
from .models_keras import FuJiangmeng
from .models_keras import ShiHaotian
from .models_keras import HuangMeiLing
from .models_keras import LihOhShu
from .models_keras import GaoJunLi
from .models_keras import WeiXiaoyan
from .models_keras import KongZhengmin
from .models_keras import YildirimOzal
from .models_keras import CaiWenjuan
from .models_keras import KimMinGu
from .models_keras import HtetMyetLynn
from .models_keras import ZhangJin
from .models_keras import YaoQihang
from .models_keras import YiboGao
from .models_keras import HongTan
from .models_keras import SharPar
from .models_keras import DaiXiLi
from .models_pytorch import TSFEDL_BaseModule
from .models_pytorch import OhShuLih
from .models_pytorch import OhShuLih_Classifier
from .models_pytorch import KhanZulfiqar
from .models_pytorch import KhanZulfiqar_Classifier
from .models_pytorch import ZhengZhenyu
from .models_pytorch import ZhengZhenyu_Classifier
from .models_pytorch import WangKejun
from .models_pytorch import WangKejun_Classifier
from .models_pytorch import ChenChen
from .models_pytorch import ChenChen_Classifier
from .models_pytorch import KimTaeYoung
from .models_pytorch import KimTaeYoung_Classifier
from .models_pytorch import GenMinxing
from .models_pytorch import GenMinxing_Classifier
from .models_pytorch import FuJiangmeng
from .models_pytorch import FuJiangmeng_Classifier
from .models_pytorch import ShiHaotian
from .models_pytorch import ShiHaotian_Classifier
from .models_pytorch import HuangMeiLing
from .models_pytorch import HuangMeiLing_Classifier
from .models_pytorch import LihOhShu
from .models_pytorch import LiOhShu_Classifier
from .models_pytorch import GaoJunLi
from .models_pytorch import GaoJunLi_Classifier
from .models_pytorch import WeiXiaoyan
from .models_pytorch import WeiXiaoyan_Classifier
from .models_pytorch import KongZhengmin
from .models_pytorch import KongZhengmin_Classifier
from .models_pytorch import YildirimOzal
from .models_pytorch import CaiWenjuan
from .models_pytorch import HtetMyetLynn
from .models_pytorch import ZhangJin
from .models_pytorch import ZhangJin_Classifier
from .models_pytorch import YaoQihang
from .models_pytorch import YaoQihangClassifier
from .models_pytorch import YiboGao
from .models_pytorch import YiboGaoClassifier
from .models_pytorch import HongTan
from .models_pytorch import HongTan_Classifier
from .models_pytorch import SharPar
from .models_pytorch import SharPar_Classifier
from .models_pytorch import DaiXiLi
from .models_pytorch import DaiXiLi_Classifier
from .utils import TimeDistributed
from .utils import flip_indices_for_conv_to_lstm
from .utils import flip_indices_for_conv_to_lstm_reshape
from .utils import check_inputs
from .utils import full_convolution

__all__ = ['densenet_transition_block',
            'densenet_conv_block',
            'densenet_dense_block',
            'squeeze_excitation_module',
            'conv_block_YiboGao',
            'attention_branch_YiboGao',
            'RTA_block',
            'spatial_attention_block_ZhangJin',
            'temporal_attention_block_ZhangJin',
            'ConvBlockYiboGao',
            'AttentionBranchYiboGao',
            'RTABlock',
            'SqueezeAndExcitationModule',
            'DenseNetTransitionBlock',
            'DenseNetConvBlock',
            'DenseNetDenseBlock',
            'SpatialAttentionBlockZhangJin',
            'TemporalAttentionBlockZhangJin',
            'get_mit_bih_segments',
            'read_mit_bih',
            'MIT_BIH',
            'OhShuLih',
            'KhanZulfiqar',
            'ZhengZhenyu',
            'HouBoroui',
            'WangKejun',
            'ChenChen',
            'KimTaeYoung',
            'GenMinxing',
            'FuJiangmeng',
            'ShiHaotian',
            'HuangMeiLing',
            'LihOhShu',
            'GaoJunLi',
            'WeiXiaoyan',
            'KongZhengmin',
            'YildirimOzal',
            'CaiWenjuan',
            'KimMinGu',
            'HtetMyetLynn',
            'ZhangJin',
            'YaoQihang',
            'YiboGao',
            'HongTan',
            'SharPar',
            'DaiXiLi',
            'TSFEDL_BaseModule',
            'OhShuLih',
            'OhShuLih_Classifier',
            'KhanZulfiqar',
            'KhanZulfiqar_Classifier',
            'ZhengZhenyu',
            'ZhengZhenyu_Classifier',
            'WangKejun',
            'WangKejun_Classifier',
            'ChenChen',
            'ChenChen_Classifier',
            'KimTaeYoung',
            'KimTaeYoung_Classifier',
            'GenMinxing',
            'GenMinxing_Classifier',
            'FuJiangmeng',
            'FuJiangmeng_Classifier',
            'ShiHaotian',
            'ShiHaotian_Classifier',
            'HuangMeiLing',
            'HuangMeiLing_Classifier',
            'LihOhShu',
            'LiOhShu_Classifier',
            'GaoJunLi',
            'GaoJunLi_Classifier',
            'WeiXiaoyan',
            'WeiXiaoyan_Classifier',
            'KongZhengmin',
            'KongZhengmin_Classifier',
            'YildirimOzal',
            'CaiWenjuan',
            'HtetMyetLynn',
            'ZhangJin',
            'ZhangJin_Classifier',
            'YaoQihang',
            'YaoQihangClassifier',
            'YiboGao',
            'YiboGaoClassifier',
            'HongTan',
            'HongTan_Classifier',
            'SharPar',
            'SharPar_Classifier',
            'DaiXiLi',
            'DaiXiLi_Classifier',
            'TimeDistributed',
            'flip_indices_for_conv_to_lstm',
            'flip_indices_for_conv_to_lstm_reshape',
            'check_inputs',
            'full_convolution']



import sys
sys.path.append("..")
sys.path.append(".")
import os

import unittest
import torch
import torch.nn as nn
import torchmetrics
import pytorch_lightning as pl
from TSFEDL.data import MIT_BIH
from torch.utils.data.sampler import SubsetRandomSampler
from TSFEDL.models_pytorch import *
import inspect


def acc_from_logits(y_hat, y):
    y_hat = F.softmax(y_hat, dim=1)
    preds = torch.argmax(y_hat, dim=1)
    acc = (preds == y).sum().item() / len(y)
    return acc


class MyTestCase(unittest.TestCase):

    def setUp(self) -> None:
        pl.seed_everything(42)
        if not os.path.isdir("physionet.org/files/mitdb/1.0.0/"):
            os.system("wget -r -N -c -np https://physionet.org/files/mitdb/1.0.0/")

    def trainModel(self, model, data_length, epochs, hot_coded=False, evaluate_test=True):
        if torch.cuda.is_available():
            torch.cuda.empty_cache()    
        mit_bih = MIT_BIH(path="physionet.org/files/mitdb/1.0.0/", return_hot_coded=hot_coded)
        mit_bih.x = mit_bih.x[:10]
        mit_bih.y = mit_bih.y[:10]
        tra_size = int(len(mit_bih) * 0.8)
        tst_size = len(mit_bih) - tra_size
        train, test = torch.utils.data.random_split(mit_bih, [tra_size, tst_size])
        train_loader = torch.utils.data.DataLoader(train, batch_size=1, num_workers=0)
        test_loader = torch.utils.data.DataLoader(test, batch_size=1, num_workers=0)
        # train_sampler = SubsetRandomSampler(range(len(mit_bih)))
        # train_loader = torch.utils.data.DataLoader(mit_bih, batch_size=256, sampler=train_sampler, num_workers=8)


        # Train
        trainer = None
        if torch.cuda.is_available():
            trainer = pl.Trainer(gpus=1, max_epochs=epochs)
        else:
            trainer = pl.Trainer(max_epochs=epochs)
        trainer.fit(model, train_loader)
        if evaluate_test:
            test_results = None
            if "model" in inspect.getfullargspec(trainer.test).args:
                test_results = trainer.test(model, test_loader)
            else:
                test_results = trainer.test(test_loader)
            return test_results[0]['test_acc_epoch']

    #@unittest.skip
    def test_LiOhShu(self):
        model = LihOhShu(in_features=1,
                         loss=nn.CrossEntropyLoss(),
                         metrics={'acc': acc_from_logits},
                         optimizer=torch.optim.Adam,
                         lr=0.001)

        acc = self.trainModel(model, 2000, 1)
        assert 1.0 >= acc >= 0

    #@unittest.skip
    def test_OhShuLi(self):
        model = OhShuLih(in_features=1,
                         loss=nn.CrossEntropyLoss(),
                         metrics={"acc": acc_from_logits},
                         optimizer=torch.optim.Adam,
                         lr=0.001)

        acc = self.trainModel(model, 1000, 1)
        assert 1.0 >= acc >= 0


    # def test_YiboGao(self):  # TODO: Test more carefully
    #     model = YiboGao(in_features=1,
    #                      metrics={"acc": acc_from_logits},
    #                      optimizer=torch.optim.Adam,
    #                      lr=0.001)
    #
    #     acc = self.trainModel(model, 1000, 10, True)
    #     assert 1.0 >= acc >= 0

    #@unittest.skip
    def test_YaoQihang(self):
        model = YaoQihang(in_features=1,
                          loss=nn.CrossEntropyLoss(),
                          metrics={"acc": acc_from_logits},
                          optimizer=torch.optim.Adam,
                          lr=0.001
                          )
        acc = self.trainModel(model, 250, 1)
        assert 1.0 >= acc >= 0

    #@unittest.skip
    def test_HtetMyetLynn(self):
        model = HtetMyetLynn(in_features=1,
                          loss=nn.CrossEntropyLoss(),
                          metrics={"acc": acc_from_logits},
                          optimizer=torch.optim.Adam,
                          lr=0.001
                          )
        acc = self.trainModel(model, 750, 1)
        assert 1.0 >= acc >= 0

    #@unittest.skip
    def test_YildirimOzal(self):
        model = YildirimOzal(
            input_shape=(1,1000),
            top_module=nn.Linear(32, 5),
            loss=nn.MSELoss(),
            metrics={"acc": acc_from_logits},
            optimizer=torch.optim.Adam,
            lr=0.001
        )
        # First, train the autoencoder
        self.trainModel(model, 260, 1, False, False)

        # Next, train the LSTM and the classifier with the encoded features.
        model.train_autoencoder = False
        model.encoder.requires_grad_(False)
        model.decoder.requires_grad_(False)
        model.loss = nn.CrossEntropyLoss()
        acc = self.trainModel(model, 1000, 1, False, True)

        assert 1.0 >= acc >= 0

    #####################################################

    #@unittest.skip
    def test_KhanZulfiqar(self):
        model = KhanZulfiqar(in_features=1,
                          loss=nn.CrossEntropyLoss(),
                          metrics={"acc": acc_from_logits},
                          optimizer=torch.optim.Adam,
                          lr=0.001
                          )
        acc = self.trainModel(model, 1000, 1)
        assert 1.0 >= acc >= 0

    #@unittest.skip
    def test_ZhengZhenyu(self):
        model = ZhengZhenyu(in_features=1,
                          loss=nn.CrossEntropyLoss(),
                          metrics={"acc": acc_from_logits},
                          optimizer=torch.optim.Adam,
                          lr=0.001
                          )
        acc = self.trainModel(model, 128, 1)
        assert 1.0 >= acc >= 0

    # # No funciona
     ###@unittest.skip
    # def test_HouBoroui(self):
    #     model = HouBoroui(
    #         in_features=1,
    #         top_module=nn.Linear(1, 5),
    #         loss=nn.MSELoss(),
    #         metrics={"acc": acc_from_logits},
    #         optimizer=torch.optim.Adam,
    #         lr=0.001
    #     )
    #     # First, train the autoencoder
    #     self.trainModel(model, 1000, 30, False, False)
    #
    #     # Next, train the LSTM and the classifier with the encoded features.
    #     model.train_autoencoder = False
    #     model.encoder.requires_grad_(False)
    #     model.decoder.requires_grad_(False)
    #     model.loss = nn.CrossEntropyLoss()
    #     acc = self.trainModel(model, 1000, 30, False, True)
    #
    #     assert 1.0 >= acc >= 0

    #@unittest.skip
    def test_WangKejun(self):
        model = WangKejun(in_features=1,
                          loss=nn.CrossEntropyLoss(),
                          metrics={"acc": acc_from_logits},
                          optimizer=torch.optim.Adam,
                          lr=0.001
                          )
        acc = self.trainModel(model, 100, 1)
        assert 1.0 >= acc >= 0

    #@unittest.skip
    def test_ChenChen(self):
        model = ChenChen(in_features=1,
                          loss=nn.CrossEntropyLoss(),
                          metrics={"acc": acc_from_logits},
                          optimizer=torch.optim.Adam,
                          lr=0.001
                          )
        acc = self.trainModel(model, 3600, 1)
        assert 1.0 >= acc >= 0

    #@unittest.skip
    def test_KimTaeYoung(self):
        model = KimTaeYoung(in_features=1,
                          loss=nn.CrossEntropyLoss(),
                          metrics={"acc": acc_from_logits},
                          optimizer=torch.optim.Adam,
                          lr=0.001
                          )
        acc = self.trainModel(model, 100, 1)
        assert 1.0 >= acc >= 0

    #@unittest.skip
    def test_GenMinxing(self):
        model = GenMinxing(in_features=1000,
                          loss=nn.CrossEntropyLoss(),
                          metrics={"acc": acc_from_logits},
                          optimizer=torch.optim.Adam,
                          lr=0.001
                          )
        acc = self.trainModel(model, 100, 1)
        assert 1.0 >= acc >= 0

    #@unittest.skip
    def test_FuJiangmeng(self):
        model = FuJiangmeng(in_features=1,
                          loss=nn.CrossEntropyLoss(),
                          metrics={"acc": acc_from_logits},
                          optimizer=torch.optim.Adam,
                          lr=0.001
                          )
        acc = self.trainModel(model, 100, 1)
        assert 1.0 >= acc >= 0

    #@unittest.skip
    def test_ShiHaotian(self):
        model = ShiHaotian(in_features=1,
                          loss=nn.CrossEntropyLoss(),
                          metrics={"acc": acc_from_logits},
                          optimizer=torch.optim.Adam,
                          lr=0.001
                          )
        acc = self.trainModel(model, 100, 1)
        assert 1.0 >= acc >= 0

    #@unittest.skip
    def test_HuangMeiLing(self):
        model = HuangMeiLing(in_features=1,
                          loss=nn.CrossEntropyLoss(),
                          metrics={"acc": acc_from_logits},
                          optimizer=torch.optim.Adam,
                          lr=0.001
                          )
        acc = self.trainModel(model, 100, 1)
        assert 1.0 >= acc >= 0

    #@unittest.skip
    def test_SharPar(self):
        model = SharPar(in_features=1,
                          loss=nn.CrossEntropyLoss(),
                          metrics={"acc": acc_from_logits},
                          optimizer=torch.optim.Adam,
                          lr=0.001
                          )
        acc = self.trainModel(model, 100, 1)
        assert 1.0 >= acc >= 0

    #@unittest.skip
    def test_DaiXiLi(self):
        model = DaiXiLi(in_features=1,
                          loss=nn.CrossEntropyLoss(),
                          metrics={"acc": acc_from_logits},
                          optimizer=torch.optim.Adam,
                          lr=0.001
                          )
        acc = self.trainModel(model, 100, 1)
        assert 1.0 >= acc >= 0


if __name__ == '__main__':
    unittest.main()



import sys
sys.path.append("..")
sys.path.append(".")
import os
import random as rn
import unittest

import numpy as np
import tensorflow as tf
from sklearn.metrics import accuracy_score, mean_squared_error
from sklearn.model_selection import train_test_split
from tensorflow import keras

import TSFEDL.models_keras as TSFEDL
from TSFEDL.data import read_mit_bih


class TestMethods(unittest.TestCase):

    def setUp(self) -> None:
        # https://keras.io/getting_started/faq/#how-can-i-obtain-reproducible-results-using-keras-during-development
        os.environ['PYTHONHASHSEED'] = '0'

        # The below is necessary for starting Numpy generated random numbers
        # in a well-defined initial state.
        np.random.seed(42)

        # The below is necessary for starting core Python generated random numbers
        # in a well-defined state.
        rn.seed(12345)

        # The below tf.set_random_seed() will make random number generation
        # in the TensorFlow backend have a well-defined initial state.
        # For further details, see: https://www.tensorflow.org/api_docs/python/tf/set_random_seed
        tf.random.set_seed(1234)
        if not os.path.isdir("physionet.org/files/mitdb/1.0.0/"):
            os.system("wget -r -N -c -np https://physionet.org/files/mitdb/1.0.0/")

    def getData(self, size=1000, test_size=0.2):
        """Gets the segments with the specified size from the MIT-BIH arrythmia dataset.
            Returns a partition of the data as follow:
            (X_tra, y_tra, y_hotcoded_tra) -> the instances and labels of the training set
            (X_tst, y_tst, y_hotcoded_tst) -> the instances and labels of the test set.
        """
        print("Load data...")
        dir = "physionet.org/files/mitdb/1.0.0/"
        X, y = read_mit_bih(dir, fixed_length=size)
        y_hot_encoded = np.zeros((y.size, y.max() + 1))
        y_hot_encoded[np.arange(y.size), y] = 1
        X_tra, X_tst, y_tra, y_tst, y_HC_tra, y_HC_tst = train_test_split(X, y, y_hot_encoded,
                                                                          test_size=test_size,
                                                                          random_state=1234,
                                                                          stratify=y)
        return (X_tra[:1], y_tra[:1], y_HC_tra[:1]), (X_tst[:1], y_tst[:1], y_HC_tst[:1])

    def trainModel(self, model, X_tra, y_tra, y_hot_encoded_tra,
                   X_tst, y_tst, y_hot_encoded_tst,
                   loss=tf.keras.losses.categorical_crossentropy,
                   metrics=['accuracy'],
                   batch_size=256,
                   epochs=6):
        model.compile(optimizer='Adam', loss=loss, metrics=metrics)
        model.fit(X_tra, y_hot_encoded_tra, batch_size=batch_size, epochs=epochs, verbose=1)

        # evaluate
        pred = np.argmax(model.predict(X_tst), axis=1)
        acc = accuracy_score(y_tst, pred)
        print("Accuracy: " + str(acc))
        return acc, pred

    #@unittest.skip
    def test_YiboGao(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=500)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model, loss = TSFEDL.YiboGao(input_tensor=input, include_top=True, return_loss=True)
        model.summary()

        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                           loss=loss,
                                           epochs=1)

        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('YiboGao_test.h5')

        # re-load de model
        new_model = TSFEDL.YiboGao(input_tensor=input, weights='YiboGao_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('YiboGao_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_YaoQihang(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=250)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))

        # The number units, filters, kernel_sizes, etc. in this model are not specified in the original paper.
        model = TSFEDL.YaoQihang(input_tensor=input, include_top=True)
        model.summary()

        # Train the model
        # If the predictions are ok, the sum of predictions should be 1: test it
        model.compile(optimizer='Adam', loss=tf.keras.losses.categorical_crossentropy, metrics=['accuracy'])
        model.fit(X_tra, y_hc_tra, batch_size=256, epochs=1, verbose=0)
        p = model.predict(X_tst)
        np.testing.assert_almost_equal(np.sum(p, axis=1), np.ones((X_tst.shape[0],)), decimal=6)

        # Once check, train the model as usual
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                           batch_size=256, epochs=1)
        print(acc)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('YaoQihang_test.h5')

        # re-load de model
        new_model = TSFEDL.YaoQihang(input_tensor=input, weights='YaoQihang_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('YaoQihang_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_ZhangJin(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=250)
        input = tf.keras.Input((15360, 12))
        model = TSFEDL.ZhangJin(input_tensor=input, include_top=True)

        # The output shapes matches the description of the paper at Section 5.2
        model.summary()

        # Now, train the model with our data.
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.ZhangJin(input_tensor=input, include_top=True)

        # Train the model
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                           batch_size=256, epochs=1)
        print(acc)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('ZhangJin_test.h5')

        # re-load de model
        new_model = TSFEDL.ZhangJin(input_tensor=input, weights='ZhangJin_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('ZhangJin_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_HtetMyetLynn(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=750)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.HtetMyetLynn(input_tensor=input, include_top=True)
        model.summary()

        # assert that output shapes are the same that in Figure 5 of the paper.
        model_output_shapes = [l.output_shape for l in model.layers
                               if not isinstance(l, tf.keras.layers.Dropout)
                               if not isinstance(l, tf.keras.layers.Conv1D)]
        desired_output_shape = [[(None, 750, 1)],
                                (None, 375, 30),
                                (None, 187, 30),
                                (None, 93, 60),
                                (None, 46, 60),
                                (None, 80),
                                (None, 5)]
        self.assertTrue(model_output_shapes == desired_output_shape,
                        msg="Output shapes do not match. Model shapes: " + str(model_output_shapes))

        # Train the model
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                           batch_size=256, epochs=1)
        print(acc)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('HtetMyetLynn_test.h5')

        # re-load de model
        new_model = TSFEDL.HtetMyetLynn(input_tensor=input, weights='HtetMyetLynn_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('HtetMyetLynn_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_CaiWenjuann(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=100)
        input = tf.keras.Input((2160, 12))  # The shape of the paper's data.

        # The number units, filters, kernel_sizes, etc. in this model are not specified in the original paper.
        model = TSFEDL.CaiWenjuan(input_tensor=input, include_top=True, classes=2)
        model.summary()
        # In the paper, for 12-variables ECG data, only the number of paramwters of the final model are given.
        # The number of params in this case are 69087. Therefore, we assert that the number of params is almost the same
        assert 70000 >= model.count_params() >= 68500

        # Once checked that model's parameters are almost the same as in the original paper. We train the model with
        # our data.
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.CaiWenjuan(input_tensor=input, include_top=True)

        # Train the model
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                           batch_size=256, epochs=1)
        print(acc)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('CaiWenjuan_test.h5')

        # re-load de model
        new_model = TSFEDL.CaiWenjuan(input_tensor=input, weights='CaiWenjuan_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('CaiWenjuan_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_KimMinGu(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=100)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))

        # The number units, filters, kernel_sizes, etc. in this model are not specified in the original paper.
        models = TSFEDL.KimMinGu(input_tensor=input, include_top=True)
        preds = []
        for i, model in enumerate(models):
            model.summary()

            # Train the model
            acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                               batch_size=256, epochs=1)
            assert 1.0 >= acc >= 0
            preds.append(predictions)

            # save weights
            model.save_weights("KimMinGu_test_" + str(i) + ".h5")

        # re-load model
        w = ["KimMinGu_test_" + str(i) + ".h5" for i in range(6)]
        models = TSFEDL.KimMinGu(input_tensor=input, weights=w)
        for i, model in enumerate(models):
            new_predictions = np.argmax(model.predict(X_tst), axis=1)
            np.testing.assert_allclose(preds[i], new_predictions, rtol=1e-6, atol=1e-6)
            os.remove(w[i])

        return True

    #@unittest.skip
    def test_YildirimOzal(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=260)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        ae, enc, classifier = TSFEDL.YildirimOzal(input_tensor=input, include_top=True)
        ae.summary()
        enc.summary()
        classifier.summary()

        # assert that output shapes are the same that in Table 3 of the paper.
        model_output_shapes = [l.output_shape for l in ae.layers
                               if not isinstance(l, tf.keras.layers.Dropout)]
        desired_output_shape = [[(None, 260, 1)],
                                (None, 260, 16),
                                (None, 130, 16),
                                (None, 130, 64),
                                (None, 130, 64),
                                (None, 65, 64),
                                (None, 65, 32),
                                (None, 65, 1),
                                (None, 32, 1),
                                (None, 32, 1),
                                (None, 32, 32),
                                (None, 64, 32),
                                (None, 64, 64),
                                (None, 128, 64),
                                (None, 128, 16),
                                (None, 2048),
                                (None, 260)]
        self.assertTrue(model_output_shapes == desired_output_shape,
                        msg="Output shapes do not match. Model shapes: " + str(model_output_shapes))

        # Train the autoencoder
        ae.compile(optimizer='Adam', loss=tf.keras.losses.MSE, metrics=['mse'])
        ae.fit(X_tra, X_tra, batch_size=256, epochs=1, verbose=0)
        preds = ae.predict(X_tst)
        # Compute average MSE
        avg_mse_per_instance = [mean_squared_error(true, pred) for true, pred in zip(X_tst, preds)]
        avg_mse = np.mean(avg_mse_per_instance)
        print("Average MSE: ", avg_mse)


        # Train the LSTM (mark the encoder part as non-trainable)
        enc.trainable = False
        ae.trainable = False
        acc, class_preds = self.trainModel(classifier, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst, epochs=1)
        # classifier.compile(optimizer='Adam', loss=tf.keras.losses.categorical_crossentropy, metrics=['accuracy'])
        # classifier.fit(X_tra, y_hc_tra, batch_size=256, epochs=30, verbose=0)
        # class_preds = np.argmax(classifier.predict(X_tst), axis=1)
        # acc = accuracy_score(y_tst, class_preds)
        # print("Accuracy: ", acc)

        # Assert test accuracy above 90%
        assert 1.0 >= acc >= 0

        # save weights
        ae.save_weights('YildirimOzal_test.h5')
        classifier.save_weights('YildirimOzal_classifier_test.h5')

        # Assert testing prediction
        # re-load de models
        ae, enc, classifier = TSFEDL.YildirimOzal(input_tensor=input,
                                                      autoencoder_weights='YildirimOzal_test.h5',
                                                      lstm_weights='YildirimOzal_classifier_test.h5')
        new_predictions = ae.predict(X_tst)
        np.testing.assert_allclose(preds, new_predictions, rtol=1e-6, atol=1e-6)

        new_predictions = np.argmax(classifier.predict(X_tst), axis=1)
        os.remove('YildirimOzal_test.h5')
        os.remove('YildirimOzal_classifier_test.h5')
        return np.testing.assert_allclose(class_preds, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_KongZhengmin(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=100)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))

        # The number units, filters, kernel_sizes, etc. in this model are not specified in the original paper.
        model = TSFEDL.KongZhengmin(input_tensor=input, include_top=True)
        model.summary()

        # Train the model
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                           batch_size=256, epochs=1)
        print(acc)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('KongZhengmin_test.h5')

        # re-load de model
        new_model = TSFEDL.KongZhengmin(input_tensor=input, weights='KongZhengmin_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('KongZhengmin_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_WeiXiaoyan(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=100)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.WeiXiaoyan(input_tensor=input, include_top=True)
        model.summary()  # Description of layers available at table 2 of the paper. No output shapes nor nº of params available

        # Train the model
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                           batch_size=256, epochs=1)
        print(acc)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('WeiXiaoyan_test.h5')

        # re-load de model
        new_model = TSFEDL.WeiXiaoyan(input_tensor=input, weights='WeiXiaoyan_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('WeiXiaoyan_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_GaoJunLi(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=100)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.GaoJunLi(input_tensor=input, include_top=True)
        model.summary()  # No parameters, layer descriptions or output shapes in the paper...

        # Train the model
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                           batch_size=256, epochs=1)
        print(acc)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('GaoJunLi_test.h5')

        # re-load de model
        new_model = TSFEDL.GaoJunLi(input_tensor=input, weights='GaoJunLi_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('GaoJunLi_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_LihOhShu(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=2000)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.LihOhShu(input_tensor=input, include_top=True)
        model.summary()  # No parameters, layer descriptions or output shapes in the paper...

        # assert that output shapes are the same that in Table 3 of the paper.
        model_output_shapes = [l.output_shape for l in model.layers
                               if not isinstance(l, tf.keras.layers.Dropout)]
        desired_output_shape = [[(None, 2000, 1)],
                                (None, 1981, 3),
                                (None, 990, 3),
                                (None, 981, 6),
                                (None, 490, 6),
                                (None, 486, 6),
                                (None, 243, 6),
                                (None, 239, 6),
                                (None, 119, 6),
                                (None, 110, 6),
                                (None, 55, 6),
                                (None, 10),
                                (None, 8),
                                (None, 8),
                                (None, 5)]
        self.assertTrue(model_output_shapes == desired_output_shape,
                        msg="Output shapes do not match. Model shapes: " + str(model_output_shapes))

        # Train the model
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                           batch_size=256, epochs=1)
        print(acc)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('LihOhShu_test.h5')

        # re-load de model
        new_model = TSFEDL.LihOhShu(input_tensor=input, weights='LihOhShu_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('LihOhShu_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_HuangMeiLing(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=100)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.HuangMeiLing(input_tensor=input, include_top=True)
        model.summary()  # No parameters, layer descriptions or output shapes in the paper...

        # Train the model
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                           batch_size=256, epochs=1)
        print(acc)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('HuangMeiLing_test.h5')

        # re-load de model
        new_model = TSFEDL.HuangMeiLing(input_tensor=input, weights='HuangMeiLing_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('HuangMeiLing_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_ShiHaotian(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=100)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.ShiHaotian(input_tensor=input, include_top=True)
        model.summary()  # No parameters, layer descriptions or output shapes in the paper...

        # Train the model
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                           batch_size=256, epochs=1)
        print(acc)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('ShiHaotian_test.h5')

        # re-load de model
        new_model = TSFEDL.ShiHaotian(input_tensor=input, weights='ShiHaotian_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('ShiHaotian_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_FuJiangmeng(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=100)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.FuJiangmeng(input_tensor=input, include_top=True)
        model.summary()  # No parameters, layer descriptions or output shapes in the paper...

        # Train the model
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                           batch_size=256, epochs=1)
        print(acc)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('FuJiangmeng_test.h5')

        # re-load de model
        new_model = TSFEDL.FuJiangmeng(input_tensor=input, weights='FuJiangmeng_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('FuJiangmeng_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_GenMinxing(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=100)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.GenMinxing(input_tensor=input, include_top=True)
        model.summary()  # No parameters, layer descriptions or output shapes in the paper...

        # Train the model
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                           batch_size=256, epochs=1)
        print(acc)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('GenMinxing_test.h5')

        # re-load de model
        new_model = TSFEDL.GenMinxing(input_tensor=input, weights='GenMinxing_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('GenMinxing_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_ChenChen(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=3600)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.ChenChen(input_tensor=input, include_top=True)
        model.summary()

        # assert that output shapes are the same that in Table 2 of the paper.
        model_output_shapes = [l.output_shape for l in model.layers
                               if not isinstance(l, tf.keras.layers.Flatten)]
        desired_output_shape = [[(None, 3600, 1)],
                                (None, 3350, 5),
                                (None, 1675, 5),
                                (None, 1526, 5),
                                (None, 763, 5),
                                (None, 664, 10),
                                (None, 332, 10),
                                (None, 252, 20),
                                (None, 126, 20),
                                (None, 66, 20),
                                (None, 33, 20),
                                (None, 20, 10),
                                (None, 10, 10),
                                (None, 10, 32),
                                (None, 10, 64),
                                (None, 5)]
        self.assertTrue(model_output_shapes == desired_output_shape,
                        msg="Output shapes do not match. Model shapes: " + str(model_output_shapes))

        # Train the model
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                           batch_size=256, epochs=1)
        print(acc)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('ChenChen_test.h5')

        # re-load de model
        new_model = TSFEDL.ChenChen(input_tensor=input, weights='ChenChen_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('ChenChen_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_KimTaeYoung(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=100)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.KimTaeYoung(input_tensor=input, include_top=True)
        model.summary()

        # Check if params matches the description in the paper at table 2
        # Note: In the original paper, the time-series is windowed. Therefore, a TimeDistributed layer is employed
        # to traverse all the generated windows. Here, we do not implement this as it is problem-specific.
        # If you need this layer, you should manually add it.
        params = [p.count_params() for p in model.layers]
        desired_params = [0, 192, 0, 8256, 0, 33024, 2080, 165]
        self.assertTrue(params == desired_params,
                        msg="Number of parameters do not match. Model params: " + str(params))
        # Train the model
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra, X_tst, y_tst, y_hc_tst,
                                           batch_size=256, epochs=1)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('KimTaeYoung_test.h5')

        # re-load de model
        new_model = TSFEDL.KimTaeYoung(input_tensor=input, weights='KimTaeYoung_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('KimTaeYoung_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_WangKejun(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=100)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.WangKejun(input_tensor=input, include_top=True)
        model.summary()

        # Train the model
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra,
                                           X_tst, y_tst, y_hc_tst, batch_size=256, epochs=1)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('WangKejun_test.h5')

        # re-load de model
        new_model = TSFEDL.WangKejun(input_tensor=input, weights='WangKejun_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('WangKejun_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_HouBoroui(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=100)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model, encoder = TSFEDL.HouBoroui(input_tensor=input)
        model.summary()  # Layer shapes not available in the paper.

        # Train the model
        model.compile(optimizer='Adam', loss='mse', metrics=['mse'])
        model.fit(X_tra, X_tra, batch_size=256, epochs=1, verbose=0)
        predictions = model.predict(X_tst)

        # Compute average MSE
        avg_mse_per_instance = [mean_squared_error(true, pred) for true, pred in zip(X_tst, predictions)]
        avg_mse = np.mean(avg_mse_per_instance)
        print("Average MSE: ", avg_mse)
        assert 1 >= avg_mse

        # save weights
        model.save_weights('HouBoroui_test.h5')

        # re-load de model
        new_model, encoder = TSFEDL.HouBoroui(input_tensor=input, weights='HouBoroui_test.h5')
        new_predictions = new_model.predict(X_tst)
        os.remove('HouBoroui_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_ZhengZhenyu(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=128)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.ZhengZhenyu(input_tensor=input, include_top=True)
        model.summary()

        # assert that output shapes are the same that in Table 2 of the paper.
        model_output_shapes = [l.output_shape for l in model.layers
                               if not isinstance(l, tf.keras.layers.BatchNormalization)
                               if not isinstance(l, tf.keras.layers.Dropout)]
        desired_output_shape = [[(None, 128, 1)],
                                (None, 128, 64),  (None, 128, 64), (None, 64, 64),
                                (None, 64, 128), (None, 64, 128), (None, 32, 128),
                                (None, 32, 256), (None, 32, 256), (None, 16, 256),
                                (None, 256),
                                (None, 2048),
                                (None, 2048),
                                (None, 5)]
        self.assertTrue(model_output_shapes == desired_output_shape,
                        msg="Output shapes do not match. Model shapes: " + str(model_output_shapes))
        # Train the model
        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra,
                                           X_tst, y_tst, y_hc_tst, batch_size=256, epochs=1)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('ZhengZhenyu_test.h5')

        # re-load de model
        new_model = TSFEDL.ZhengZhenyu(input_tensor=input, weights='ZhengZhenyu_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('ZhengZhenyu_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_KhanZulfiqar(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=100)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.KhanZulfiqar(input_tensor=input, include_top=True)
        model.summary()

        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra,
                                           X_tst, y_tst, y_hc_tst, batch_size=256, epochs=1)
        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('KhanZulfiqar_test.h5')

        # re-load de model
        new_model = TSFEDL.KhanZulfiqar(input_tensor=input, weights='KhanZulfiqar_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('KhanZulfiqar_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_OhShuLih(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=1000)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.OhShuLih(input_tensor=input, include_top=True)

        # assert that output shapes are the same that in Table 2 of the paper.
        model_output_shapes = [l.output_shape for l in model.layers
                               if not isinstance(l, tf.keras.layers.ZeroPadding1D)]
        desired_output_shape = [[(None, 1000, 1)],
                                (None, 1019, 3),
                                (None, 509, 3),
                                (None, 518, 6),
                                (None, 259, 6),
                                (None, 263, 6),
                                (None, 131, 6),
                                (None, 20),
                                (None, 20),
                                (None, 20),
                                (None, 10),
                                (None, 5)]
        self.assertTrue(model_output_shapes == desired_output_shape,
                        msg="Output shapes do not match. Model shapes: " + str(model_output_shapes))

        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra,
                                           X_tst, y_tst, y_hc_tst, batch_size=256, epochs=1)

        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('OhShuLi_test.h5')

        # re-load de model
        new_model = TSFEDL.OhShuLih(input_tensor=input, weights='OhShuLi_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('OhShuLi_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_SharPar(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=1000)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.SharPar(input_tensor=input, include_top=True)

        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra,
                                           X_tst, y_tst, y_hc_tst, batch_size=256, epochs=1)

        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('SharPar_test.h5')

        # re-load de model
        new_model = TSFEDL.SharPar(input_tensor=input, weights='SharPar_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('SharPar_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)

    #@unittest.skip
    def test_DaiXiLi(self):
        (X_tra, y_tra, y_hc_tra), (X_tst, y_tst, y_hc_tst) = self.getData(size=1000)
        input = tf.keras.Input((X_tra.shape[1], X_tra.shape[2]))
        model = TSFEDL.DaiXiLi(input_tensor=input, include_top=True)

        acc, predictions = self.trainModel(model, X_tra, y_tra, y_hc_tra,
                                           X_tst, y_tst, y_hc_tst, batch_size=256, epochs=1)

        assert 1.0 >= acc >= 0

        # save weights
        model.save_weights('DaiXiLi_test.h5')

        # re-load de model
        new_model = TSFEDL.DaiXiLi(input_tensor=input, weights='DaiXiLi_test.h5')
        new_predictions = np.argmax(new_model.predict(X_tst), axis=1)
        os.remove('DaiXiLi_test.h5')
        return np.testing.assert_allclose(predictions, new_predictions, rtol=1e-6, atol=1e-6)


if __name__ == "__main__":
    unittest.main()



import os
from typing import Optional, Tuple
import numpy as np
import wfdb
from sklearn.preprocessing import LabelEncoder
from torch.utils.data import Dataset
from sklearn.model_selection import train_test_split


def get_mit_bih_segments(data: wfdb.Record,
                         annotations: wfdb.Annotation,
                         labels: np.ndarray,
                         left_offset: int = 99,
                         right_offset: int = 160,
                         fixed_length: Optional[int] = None) -> Tuple[np.ndarray, np.ndarray]:
    """
    It generates the segments of uninterrupted sequences of arrythmia beats into the corresponding arrythmia groups
    in labels.

    Parameters
    ----------

    data : wfdb.Record
        The arrythmia signal as a wfdb Record class
    annotations : wfdb.Annotation
        The set of annotations as a wfdb Annotation class
    labels : array-like
        The set of valid labels for the different segments. Segments with different labels are discarded
    left_offset : int
        The number of instance at the left of the first R peak of the segment. Default to 99
    right_offset : int
        The number of instances at the right of the last R peak of the segment. Default to 160
    fixed_length : int, optional
        Should the segments have a fixed length? If fixed_length is a number, then the segments will
        have the specified length. If the segment length is greater than fixed_length, it is truncated
        or padded with zeros otherwise. Default to None.

    Returns
    -------
        A tuple that contains the data and the associated labels. Data has a shape of (N, T, V)
        where N is the number of segments (or instances), V is the number of variables (1 in this case)
        and T is the number of timesteps of each segment.  Labels are numerically encoded according to the
        value passed in the :parameter labels param.
    """
    i = 0
    annot_segments = []

    # Get the tuples for consecutive symbols. The tuple is (first, last, symbol) where first is the index of the first occurrence of symbol,
    # and last is the index of the last consecutive ocurrence.
    while (i < len(annotations.symbol)):
        first = i
        current_symbol = annotations.symbol[i]
        while (i < len(annotations.symbol) and annotations.symbol[i] == current_symbol):
            i += 1
        last = i - 1
        tup = (first, last, current_symbol)
        annot_segments.append(tup)

    # Now, for each extracted tuple, get the X segments:
    result = []
    classes = []
    for s in annot_segments:  # s is a tuple (first, last, symbol)
        if s[2] in labels:
            classes.append(s[2])
            init = annotations.sample[s[0]] - left_offset
            if init < 0:
                init = 0

            end = annotations.sample[s[1]] + right_offset
            if end >= len(data.p_signal):
                end = len(data.p_signal) - 1

            r = range(init, end)

            # Get the samples of the segments (p_signal is a 2D array, we only want the first axis)
            new_segment = np.array(data.p_signal[r, 1], dtype='float32')

            # truncate or pad with zeros the segment if necessary
            if (fixed_length != None):
                if (len(new_segment) > fixed_length):  # truncate
                    new_segment = new_segment[:fixed_length]
                elif (len(new_segment < fixed_length)):  # pad with zeros to the right
                    number_of_zeros = fixed_length - len(new_segment)
                    new_segment = np.pad(new_segment, (0, number_of_zeros), mode='constant', constant_values=0)

            result.append(new_segment)
    result = np.stack(result, axis=0)
    result = np.reshape(result, (result.shape[0], result.shape[1], 1))  # shape[0] segments with 1 variable, with shape[1] timestamps each
    classes = np.array(classes, dtype=str)

    # Encode labels: from string to numeric.
    label_encoder = LabelEncoder()
    label_encoder.fit(labels)
    classes = label_encoder.transform(classes)

    return (result, classes)


def read_mit_bih(path: str,
                 labels: np.ndarray = np.array(['N', 'L', 'R', 'A', 'V']),
                 left_offset: int = 99,
                 right_offset: int = 160,
                 fixed_length: Optional[int] = 1000) -> Tuple[np.ndarray, np.ndarray]:
    """
    It reads the MIT-BIH Arrythmia X with the specified default configuration of the work presented at:
    Oh, Shu Lih, et al. "Automated diagnosis of arrhythmia using combination of CNN and LSTM techniques with
    variable length heart beats." Computers in biology and medicine 102 (2018): 278-287.

    Parameters
    ----------
    labels : array-like
        The labels of the different types of arrythmia to be employed
    path : str
        The path of the directory where the X files are stored. Note: The X and annotations
        files must have the same name, but different extension (annotations must have .atr extension)
    left_offset : int
        The number of instances at the left of the first R peak of the segment. Defaults to 99
    right_offset : int
        The number of instances at the right of the last R peak of the segment. Defaults to 160
    fixed_length : int, optional
        If different to None, the segment will have the specified number of instances. Note that
        if the segment length > fixed_length it will be truncate or padded with zeros otherwise.

    Returns
    -------
        A tuple that contains the data and the associated labels as an ndarray. Data has a shape of (N, T, V)
        where N is the number of segments (or instances), V is the number of variables (1 in this case)
        and T is the number of timesteps of each segment.  Labels are numerically encoded according to the
        value passed in the :parameter labels param.
    """
    print("reading data...")
    segments = []
    classes = []

    files = [file[:-4] for file in os.listdir(path) if file.endswith('.dat')]
    for f in files:
        data = wfdb.rdrecord(path + f)
        annotation = wfdb.rdann(path + f, 'atr')

        s, clazz = get_mit_bih_segments(data=data,
                                        annotations=annotation,
                                        labels=labels,
                                        left_offset=left_offset,
                                        right_offset=right_offset,
                                        fixed_length=fixed_length)

        segments.append(s)
        classes.append(clazz)

    segments = np.vstack(segments)
    classes = np.concatenate(classes)
    print("done.")

    return (segments, classes)


class MIT_BIH(Dataset):
    """
        Reads the MIT-BIH datasets and return a data loader with Shape (N, C, L) where N is the batch size, C is the
        number of channels (1 in this dataset) and L is the `length` of the time series (1000 by default).

        Parameters
        ----------
        labels :array-like
            The labels of the different types of arrythmia to be employed
        path : str
            The path of the directory where the X files are stored. Note: The X and annotations
            files must have the same name, but different extension (annotations must have .atr extension)
        left_offset : int
            The number of instances at the left of the first R peak of the segment. Defaults to 99
        right_offset : int
            The number of instances at the right of the last R peak of the segment. Defaults to 160
        return_hot_coded : bool
            Wether to return the raw labels or hot-encoded ones.

        Returns
        -------
            A tuple that contains the data and the associated labels as an ndarray. Data has a shape of (N, T, V)
            where N is the number of segments (or instances), V is the number of variables (1 in this case)
            and T is the number of timesteps of each segment.  Labels are numerically encoded according to the
            value passed in the :parameter labels param.
    """
    def __init__(self, path,
                 labels=np.array(['N', 'L', 'R', 'A', 'V']),
                 length=1000,
                 left_offset=99,
                 right_offset=160,
                 return_hot_coded=False):
        X, y = read_mit_bih(path, labels, left_offset=left_offset, right_offset=right_offset, fixed_length=length)
        y_hot_encoded = np.zeros((y.size, y.max() + 1), dtype='int64')
        y_hot_encoded[np.arange(y.size), y] = 1

        self.x = X.reshape((X.shape[0], X.shape[2], X.shape[1]))
        if return_hot_coded:
            self.y = y_hot_encoded
        else:
            self.y = y

    def __len__(self):
        return len(self.y)

    def __getitem__(self, index):
        return self.x[index], self.y[index]
